class ImproperChapterError(Exception):
    """Exception for improper chapter files"""


class NoChapterDataError(Exception):
    """No chapter data error"""


class ChapterIndexError(Exception):
    """Chapter index error"""
